;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2019, 2020, 2021, 2022, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
;;; Copyright © 2019, 2020 Rob Wieringa <rma.wieringa@gmail.com>
;;; Copyright © 2019, 2020, 2021, 2022 Rutger van Beusekom <rutger@dezyne.org>
;;; Copyright © 2021, 2024 Paul Hoogendijk <paul@dezyne.org>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
;;;
;;; Commentary:
;;;
;;; Code:

(define-module (dzn parse peg)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-2)
  #:use-module (srfi srfi-9)
  #:use-module (srfi srfi-9 gnu)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-71)

  #:use-module (ice-9 match)

  #:use-module (dzn peg)
  #:use-module (dzn command-line)
  #:use-module (dzn misc)

  #:re-export (%peg:locations?
               %peg:skip?
               %peg:fall-back?
               %peg:debug?
               %peg:error)

  #:export (%peg:table
            peg:imports
            peg:parse
            peg:skip-parse
            peg:import-skip-parse))

(define-skip-parser peg-eof none (not-followed-by peg-any))
(define-skip-parser peg-eol none (or "\f" "\n" "\r" "\v"))
(define-skip-parser peg-ws none (or " " "\t"))
(define-skip-parser peg-line all
  (and "//" (* (and (not-followed-by peg-eol) peg-any))
       (expect (or "\n" "\r\n" peg-eof))))
(define-skip-parser peg-block-strict all
  (and "/*"
       (* (or peg-block (and (not-followed-by "*/") peg-any)))
       (expect "*/")))
(define-skip-parser peg-skip all
  (* (or peg-ws peg-eol peg-line peg-block-strict)))

(define-skip-parser peg-block all
  (and "/*" (* (or peg-block (and (not-followed-by "*/") peg-any)))
       (or "*/" peg-eof)))
(define-skip-parser peg-import-skip all
  (* (or peg-ws peg-eol peg-line peg-block)))

(define (peg:imports string)
  ;; . 'import'? eagerly SKIP anything with import in it
  (define-peg-string-patterns
    "root <- (dollars / import / SKIP)*
dollars < DOLLAR (!DOLLAR .)* DOLLAR#
file-name <- (!SEMICOLON !DOLLAR .)+
import <-- IMPORT file-name SEMICOLON
DOLLAR < '$'
IMPORT < 'import' ![a-zA-Z_0-9]
SEMICOLON < ';'
SKIP < . 'import'?")
  (peg:tree (match-pattern root string)))

(define peg:skip-parse peg-skip)
(define peg:import-skip-parse peg-import-skip)

(define-immutable-record-type <peg:symbol>
  (make-peg:symbol scope name type local offset)
  peg:symbol?
  (scope peg:symbol-scope)
  (name peg:symbol-name)
  (type peg:symbol-type)
  (local peg:symbol-local)
  (offset peg:symbol-offset))

(define %peg:table (make-parameter #f))

(define empty-symbol (string->symbol ""))

(define* (peg:parse string)
  (define %peg:symbol (make-parameter #f))
  (define %peg:current (make-parameter `((,empty-symbol))))

  (define (current:scope)
    (if (null? (%peg:current)) '()
        (reverse (car (%peg:current)))))

  (define (push-scope str len pos)
    "Record the last scoping symbol in %peg:current."
    (let ((symbol (%peg:symbol))
          (stack (%peg:current)))
      ;; symbol is either a: namespace, interface, component or frame
      (cond (symbol
             (let ((scope (peg:symbol-scope symbol))
                   (name (peg:symbol-name symbol))
                   (type (peg:symbol-type symbol)))
               (if (and (eq? type 'namespace)
                        (not (every eq? (reverse scope) (car stack))))
                   (%peg:current (cons (cons name (reverse scope)) stack))
                   (%peg:current (cons (cons name (if (null? stack) '()  (car stack))) stack))))
             ;;(pke 'push (%peg:current))
             `(,pos '()))
            (else
             `(,pos ())))))
  (define-peg-pattern push none push-scope)

  (define (pop-symbol s table)
    (cond ((and (peg:symbol-local s)
                (equal? (reverse (car (%peg:current)))
                        (peg:symbol-scope s)))
           ;;(pke 'pop (peg:symbol->string s))
           table)
          (else
           (cons s table))))

  (define (pop-scope str len pos)
    "Reduce %peg:current after removing from %peg:table all local symbols in
the same scope as %peg:current."
    (%peg:table (fold pop-symbol '() (%peg:table)))
    (when (pair? (%peg:current))
      (%peg:current (cdr (%peg:current))))
    ;;(pke 'pop (%peg:current))
    `(,pos '()))
  (define-peg-pattern pop none pop-scope)

  (define (dump-symbol-table str len pos)
    (map write-symbol (%peg:table))
    `(,pos '()))
  (define-peg-pattern dump none dump-symbol-table)

  (define (peg:symbol->string s)
    (format #f "type: ~a scope: ~a name: ~a local: ~a"
            (peg:symbol-type s)
            (peg:symbol-scope s)
            (peg:symbol-name s)
            (peg:symbol-local s)))

  (define* (write-symbol s #:optional (port (current-error-port)))
    (format port "~a\n" (peg:symbol->string s)))

  (define (global? symbol)
    (eq? empty-symbol symbol))

  (define* ((duplicate-name? scope name) symbol)
    (and (eq? name (peg:symbol-name symbol))
         (equal? scope (peg:symbol-scope symbol))))

  (define* ((symbol=? scope name type pos) symbol)
    (and (or (peg:symbol-local symbol)
             (equal? scope (peg:symbol-scope symbol)))
         (eq? name (peg:symbol-name symbol))
         (or (not type)
             (eq? type (peg:symbol-type symbol)))
         ;; duplicate when different offset
         (not (= pos (peg:symbol-offset symbol)))))

  (define* ((symbol<=? scope name type) symbol)
    (let* ((result
            (and (or (peg:symbol-local symbol)
                     (every eq? (reverse scope)
                            (reverse (peg:symbol-scope symbol))))
                 (eq? name (peg:symbol-name symbol))
                 (or (not type)
                     (eq? type (peg:symbol-type symbol))))))
      ;; (when (eq? type 'enum)
      ;;   (pke 'find '=> result)
      ;;   (pke 'type type '== (peg:symbol-type symbol))
      ;;   (pke 'scope scope '== (peg:symbol-scope symbol)
      ;;        '=> (every eq? (reverse scope) (reverse (peg:symbol-scope symbol))))
      ;;   (pke 'name name '== (peg:symbol-name symbol)))
      result))

  (define* ((set-frame-symbol type) str len pos)
    (let ((scope (current:scope)))
      (%peg:symbol (make-peg:symbol scope 'frame type #t pos))
      `(,pos ())))

  ;; a frame symbol
  (define set-frame (set-frame-symbol 'frame))
  (define-peg-pattern =frame none set-frame)

  ;; FIXME merge local and global version into add-symbol
  (define* ((add-local-symbol type) str len pos)
    "Prepare symbol for commit, unless # throws an exception.  A symbol may
be shadowed."
    (and-let* ((res (name str len pos))
               (name (string->symbol (substring str pos (car res))))
               (scope (current:scope))
               (symbol (make-peg:symbol scope name type #t pos)))
      (%peg:symbol symbol)
      (when (and (eq? type 'function)
                 (and=> (find (duplicate-name? scope name) (%peg:table))
                        (compose (cute eq? 'event <>) peg:symbol-type)))
        (throw 'syntax-error (list pos 'duplicate)))
      res))

  (define* ((add-global-symbol type) str len pos)
    "Prepare a symbol for commit.  Note that when # throws, the commit will
be skipped.  Handle namespaces with a.b.c syntax.  Namespace a and b
will not be added to the symbol table, since they are either empty or
will be added elsewhere when they are not empty."
    (and-let* ((res (compound-name str len pos)))
      (let* ((string (substring str pos (car res)))
             (prefix name (string->scope+name string))
             (current (current:scope))
             (scope (if (and (pair? prefix)
                             (eq? empty-symbol (car prefix)))
                        current
                        (append current prefix)))
             (symbol (make-peg:symbol scope name type #f pos))
             (previous? (and (not (eq? type 'namespace))
                             (find (symbol=? scope name type pos) (%peg:table)))))

        (cond ((and previous? (not (peg:symbol-local previous?)))
               ;; (pke 'new (peg:symbol->string symbol)
               ;;      'previous (peg:symbol->string previous?))
               ;; FIXME format error with previous definition location
               ;; => requires extending arguments of thrown 'syntax-error
               (throw 'syntax-error (list pos 'duplicate)))
              (else
               ;;(pke 'add-global (peg:symbol->string symbol))
               (%peg:symbol symbol)
               res)))))

  (define (commit-symbol str len pos)
    "Commit prepared symbol to table."
    (cond ((%peg:symbol)
           (%peg:table (cons (%peg:symbol) (%peg:table)))
           (%peg:symbol #f)
           `(,pos ()))
          (else
           ;;(throw 'parser-error `(,pos "no symbol"))
           `(,pos ()))))
  (define-peg-pattern commit none commit-symbol)

  (define add-function (add-local-symbol 'function))
  (define-peg-pattern =function body add-function)
  (define add-instance (add-local-symbol 'instance))
  (define-peg-pattern =instance body add-instance)
  (define add-port (add-local-symbol 'port))
  (define-peg-pattern =port body add-port)
  (define add-variable (add-local-symbol 'variable))
  (define-peg-pattern =variable body add-variable)

  (define add-enum (add-global-symbol 'enum))
  (define-peg-pattern =enum body add-enum)
  (define add-extern (add-global-symbol 'extern))
  (define-peg-pattern =extern body add-extern)
  (define add-event (add-local-symbol 'event))
  (define-peg-pattern =event body add-event)
  (define add-component (add-global-symbol 'component))
  (define-peg-pattern =component body add-component)
  (define add-interface (add-global-symbol 'interface))
  (define-peg-pattern =interface body add-interface)
  (define add-namespace (add-global-symbol 'namespace))
  (define-peg-pattern =namespace body add-namespace)
  (define add-subint (add-global-symbol 'subint))
  (define-peg-pattern =subint body add-subint)

  (define* ((local-identifier-p type) str len pos)
    (and-let* ((res (name str len pos))
               (name (string->symbol (substring str pos (car res))))
               (scope (current:scope))
               (found? (find (symbol=? scope name type 0)
                             (%peg:table))))
      ;;(pke 'local (peg:symbol->string found?))
      res))

  (define identifier-event-p (local-identifier-p 'event))
  (define-peg-pattern is-event body identifier-event-p)
  (define identifier-instance-p (local-identifier-p 'instance))
  (define-peg-pattern is-instance body identifier-instance-p)
  (define identifier-port-p (local-identifier-p 'port))
  (define-peg-pattern is-port body identifier-port-p)
  (define identifier-variable-p (local-identifier-p 'variable))
  (define-peg-pattern is-variable body identifier-variable-p)

  ;; lookup:
  ;; or
  ;; at same level of scope
  ;; at global scope

  (define (string->scope+name string)
    (match (string-split string #\.)
      ((name)
       (values `() (string->symbol name)))
      ((scope ... name)
       (values (map string->symbol scope) (string->symbol name)))))

  (define* ((global-identifier-p type) str len pos)
    (and-let* ((res (compound-name str len pos)))
      (let* ((scope name (string->scope+name (substring str pos (car res)))))
        (let* ((found? (find (symbol<=? scope name type) (%peg:table))))
          (and found? res)))))

  (define* ((enum-prefix-p type) str len pos)
    (and-let* ((res (enum-prefix str len pos)))
      (let* ((string (substring str pos (car res)))
             (string (string-trim-right string #\.))
             (scope name (string->scope+name string)))
        (let* ((found? (find (symbol<=? scope name type) (%peg:table))))
          ;;(pke 'enum-prefix 'scope scope 'name name '=> found?)
          (and found? res)))))

  (define identifier-enum-p (global-identifier-p 'enum))
  (define-peg-pattern is-enum body identifier-enum-p)
  (define identifier-enum-prefix-p (enum-prefix-p 'enum))
  (define-peg-pattern is-enum-prefix body identifier-enum-prefix-p)
  (define identifier-extern-p (global-identifier-p 'extern))
  (define-peg-pattern is-extern body identifier-extern-p)
  (define identifier-component-p (global-identifier-p 'component))
  (define-peg-pattern is-component body identifier-component-p)
  (define identifier-interface-p (global-identifier-p 'interface))
  (define-peg-pattern is-interface body identifier-interface-p)
  (define identifier-namespace-p (global-identifier-p 'namespace))
  (define-peg-pattern is-namespace body identifier-namespace-p)
  (define identifier-subint-p (global-identifier-p 'subint))
  (define-peg-pattern is-subint body identifier-subint-p)

  (define (dollars-no-skip str len pos)
    (parameterize ((%peg:skip? (lambda (str len at) `(,at ()))))
      (dollars- str len pos)))

  ;; TODO: come up with lexical scope construct (see spirit),
  ;; i.e. disable skip parser locally
  (define-peg-pattern dollars all dollars-no-skip)

  (define-peg-string-patterns
    "root <-- (import / dollars / type
            / namespace / interface / component / function
            / EOF)#*

type-name <-- is-enum / is-extern / is-subint / compound-name / BOOL / VOID

reserved <- is-event / is-variable / is-port / is-instance

import <-- IMPORT file-name SEMICOLON#
  file-name <- (!SEMICOLON .)+

dollars- <- DOLLAR (!DOLLAR .)* DOLLAR#

type <- enum / int / extern
  enum <-- ENUM =enum# BRACE-OPEN# fields# BRACE-CLOSE# commit SEMICOLON#
    fields <-- (name (&BRACE-CLOSE / COMMA#))+

  int <-- SUBINT =subint# BRACE-OPEN# range# BRACE-CLOSE# commit SEMICOLON#
    range <-- from DOTDOT# to
    from <-- NUMBER#
    to <-- NUMBER#

  extern <-- EXTERN =extern# commit dollars# SEMICOLON#

formals <-- PAREN-OPEN ((&PAREN-CLOSE / formal#) (&PAREN-CLOSE / COMMA#))* PAREN-CLOSE#
   formal <-- direction? is-interface =port# commit
            / direction? type-name =variable# commit

namespace <-- NAMESPACE =namespace push commit
              BRACE-OPEN# namespace-root pop BRACE-CLOSE#
  namespace-root <-- (namespace / interface / component / type / &BRACE-CLOSE)#*

interface <-- INTERFACE =interface# push commit
              BRACE-OPEN# types-and-events# (behavior / &BRACE-CLOSE)# pop BRACE-CLOSE#

  types-and-events <-- (type / event / &behavior / &BRACE-CLOSE)#*
    event-name <-- =event
    event <-- direction type-name# event-name# commit
              formals# SEMICOLON#
      direction <-- INOUT / IN / OUT / provides / requires

component <-- COMPONENT =component# push commit BRACE-OPEN# ports# body# pop BRACE-CLOSE#
  body <- behavior / system / &BRACE-CLOSE
    system <-- SYSTEM BRACE-OPEN# instances-and-bindings BRACE-CLOSE#
      instances-and-bindings <-- (instance / binding)*
        instance <-- (is-component / unknown-identifier) =instance SEMICOLON# commit
        binding <-- end-point BIND end-point# SEMICOLON#
          end-point <-- ASTERISK / is-port
                      / is-instance DOT (ASTERISK / name#)
                      / unknown-identifier (DOT name#)?

  ports <-- (port / &BEHAVIOR / &SYSTEM / &BRACE-CLOSE)*
    port <-- port-direction port-qualifiers
             (is-interface / unknown-identifier / EMPTY)# =port# commit SEMICOLON#
      port-direction <- provides / requires
      port-qualifiers <-- (blocking-q / external / injected)?

behavior <-- BEHAVIOR behavior-compound

  behavior-compound
    <-- BRACE-OPEN# behavior-statements BRACE-CLOSE#

    behavior-statements
       <-- (expression-function / function / variable / declarative-statement /
            type / &BRACE-CLOSE)#*

      function
         <-- type-name =function &PAREN-OPEN push commit formals compound# pop

      expression-function
         <-- type-name =function &(PAREN-OPEN PAREN-CLOSE ASSIGN) push commit
             formals ASSIGN expression# pop SEMICOLON#

declarative-statement <- on / blocking / guard / compound / invariant
  on <-- ON triggers# COLON# (statement / !unknown-identifier)#

    triggers <-- ((trigger / !unknown-identifier) (&COLON / COMMA)#)#+
      trigger <-- is-port DOT# name# (&(COLON illegal) / trigger-formals#)
                / is-event
                / OPTIONAL / INEVITABLE
        trigger-formals <-- PAREN-OPEN (formal-reference (&PAREN-CLOSE / COMMA#))*
                            PAREN-CLOSE#
          formal-reference <-- =variable (LEFT-ARROW reference)? commit

  guard
    <-- BRACKET-OPEN (otherwise / expression)# BRACKET-CLOSE# statement#

compound
   <-- BRACE-OPEN =frame push commit (statement / !unknown-identifier)#* pop BRACE-CLOSE#

  statement
     <- (declarative-statement / imperative-statement / !unknown-identifier)#

invariant
   <-- INVARIANT expression# SEMICOLON#

imperative-statement
  <- variable / assign / if-statement / illegal / return / skip-statement
     / compound / reply / defer / action SEMICOLON# / call SEMICOLON#

  defer <-- DEFER arguments? imperative-statement

  action <-- is-port DOT name# arguments / is-event arguments?
  call <-- compound-name arguments
    arguments <-- PAREN-OPEN (argument (&PAREN-CLOSE / COMMA#))* PAREN-CLOSE#
      argument <-- expression

  skip-statement <-- SEMICOLON

  blocking <-- BLOCKING statement

  illegal <-- ILLEGAL SEMICOLON# / BRACE-OPEN ILLEGAL SEMICOLON# BRACE-CLOSE#

  assign <-- reference ASSIGN expression# SEMICOLON#

  if-statement <-- IF PAREN-OPEN# expression# PAREN-CLOSE# imperative-statement#
                   (ELSE imperative-statement#)?

  reply
    <-- (name DOT)? REPLY PAREN-OPEN# expression? PAREN-CLOSE# SEMICOLON#

  return
    <-- RETURN expression? SEMICOLON#

  variable
    <-- type-name =variable (ASSIGN expression#)? commit SEMICOLON#

  reference
    <-- is-variable

  port-reference
    <-- is-port

expression
  <-- implies-expression

implies-expression
  <- or-expression IMPLIES implies-expression# / or-expression

or-expression
  <- and-expression OR or-expression# / and-expression

and-expression
  <- compare-expression AND and-expression# / compare-expression

compare-expression
  <- plus-min-expression !LEFT-ARROW COMPARE plus-min-expression#
     / plus-min-expression

plus-min-expression
  <- not-expression (PLUS / MINUS) plus-min-expression# / not-expression

not-expression
  <- minus / negation / group / dollars / literal / field-test / reference
   / action / call / shared-field-test / shared-reference
   / port-reference
   / enum-literal

negation <-- NOT not-expression#
minus <-- UMINUS not-expression#
enum-literal <-- is-enum-prefix name / unknown-identifier
field-test <-- reference DOT name#
shared-reference <-- is-port DOT name#
shared-field-test <-- shared-reference DOT name#
literal <-- NUMBER / FALSE / TRUE
group <-- PAREN-OPEN expression PAREN-CLOSE#

name <-- !KEYWORD identifier

compound-name <-- global? scope? name

enum-prefix <- global? scope? &name

scope <-- (name DOT &name)+

global <-- DOT

identifier <- [a-zA-Z_] [a-zA-Z_0-9]*

compound-identifier <- [a-zA-Z_] [a-zA-Z_0-9]* ([.] [a-zA-Z_] [a-zA-Z_0-9]*)*

unknown-identifier <-- !KEYWORD compound-identifier

otherwise <-- OTHERWISE
provides <-- PROVIDES
requires <-- REQUIRES
blocking-q <-- BLOCKING
external <-- EXTERNAL
injected <-- INJECTED

UMINUS              <  MINUS
NUMBER              <-  MINUS? [0-9]+
ASTERISK            <-  '*'
DOLLAR              <   '$'
BRACE-OPEN          <   '{'
BRACE-CLOSE         <   '}'
BRACKET-OPEN        <   '['
BRACKET-CLOSE       <   ']'
PAREN-OPEN          <   '('
PAREN-CLOSE         <   ')'
SEMICOLON           <   ';'
COLON               <   ':'
DOT                 <   '.'
DOTDOT              <   '..'
COMMA               <   ','
BIND                <   '<=>'
ASSIGN              <   '='
LEFT-ARROW          <   '<-'
OR                  <-  '||'
AND                 <-  '&&'
EQUAL               <-  '=='
NOT-EQUAL           <-  '!='
LESS                <-  '<'
LESS-EQUAL          <-  '<='
GREATER             <-  '>'
GREATER-EQUAL       <-  '>='
IMPLIES             <-  '=>'
PLUS                <-  '+'
MINUS               <-  '-'
NOT                 <   '!'
EMPTY               <   &.
EOF                 <   !.
COMPARE             <-  EQUAL / NOT-EQUAL / LESS-EQUAL / LESS
                        / GREATER-EQUAL / GREATER
BEHAVIOR            <  'behavior' ![a-zA-Z_0-9] / 'behaviour' ![a-zA-Z_0-9]
BLOCKING            <  'blocking' ![a-zA-Z_0-9]
BOOL                <- 'bool' ![a-zA-Z_0-9]
COMPONENT           <  'component' ![a-zA-Z_0-9]
DEFER               <  'defer' ![a-zA-Z_0-9]
ELSE                <  'else' ![a-zA-Z_0-9]
ENUM                <  'enum' ![a-zA-Z_0-9]
EXTERN              <  'extern' ![a-zA-Z_0-9]
EXTERNAL            <  'external' ![a-zA-Z_0-9]
FALSE               <- 'false' ![a-zA-Z_0-9]
IF                  <  'if' ![a-zA-Z_0-9]
ILLEGAL             <  'illegal' ![a-zA-Z_0-9]
IMPORT              <  'import' ![a-zA-Z_0-9]
IN                  <- 'in' ![a-zA-Z_0-9]
INEVITABLE          <- 'inevitable' ![a-zA-Z_0-9]
INJECTED            <  'injected' ![a-zA-Z_0-9]
INOUT               <- 'inout' ![a-zA-Z_0-9]
INTERFACE           <  'interface' ![a-zA-Z_0-9]
INVARIANT           <  'invariant' ![a-zA-Z_0-9]
NAMESPACE           <  'namespace' ![a-zA-Z_0-9]
ON                  <  'on' ![a-zA-Z_0-9]
OPTIONAL            <- 'optional' ![a-zA-Z_0-9]
OTHERWISE           <  'otherwise' ![a-zA-Z_0-9]
OUT                 <- 'out' ![a-zA-Z_0-9]
PROVIDES            <  'provides' ![a-zA-Z_0-9]
REPLY               <- 'reply' ![a-zA-Z_0-9]
REQUIRES            <  'requires' ![a-zA-Z_0-9]
RETURN              <  'return' ![a-zA-Z_0-9]
SUBINT              <  'subint' ![a-zA-Z_0-9]
SYSTEM              <  'system' ![a-zA-Z_0-9]
TRUE                <- 'true' ![a-zA-Z_0-9]
VOID                <- 'void' ![a-zA-Z_0-9]
NEWLINE             <- '\n'

KEYWORD <
    'behavior' ![a-zA-Z_0-9]
  / 'behaviour' ![a-zA-Z_0-9]
  / 'blocking' ![a-zA-Z_0-9]
  / 'component' ![a-zA-Z_0-9]
  / 'defer' ![a-zA-Z_0-9]
  / 'else' ![a-zA-Z_0-9]
  / 'enum' ![a-zA-Z_0-9]
  / 'extern' ![a-zA-Z_0-9]
  / 'external' ![a-zA-Z_0-9]
  / 'false' ![a-zA-Z_0-9]
  / 'if' ![a-zA-Z_0-9]
  / 'illegal' ![a-zA-Z_0-9]
  / 'import' ![a-zA-Z_0-9]
  / 'in' ![a-zA-Z_0-9]
  / 'inevitable' ![a-zA-Z_0-9]
  / 'injected' ![a-zA-Z_0-9]
  / 'inout' ![a-zA-Z_0-9]
  / 'interface' ![a-zA-Z_0-9]
  / 'invariant' ![a-zA-Z_0-9]
  / 'namespace' ![a-zA-Z_0-9]
  / 'on' ![a-zA-Z_0-9]
  / 'optional' ![a-zA-Z_0-9]
  / 'otherwise' ![a-zA-Z_0-9]
  / 'out' ![a-zA-Z_0-9]
  / 'provides' ![a-zA-Z_0-9]
  / 'reply' ![a-zA-Z_0-9]
  / 'requires' ![a-zA-Z_0-9]
  / 'return' ![a-zA-Z_0-9]
  / 'subint' ![a-zA-Z_0-9]
  / 'system' ![a-zA-Z_0-9]
  / 'true' ![a-zA-Z_0-9]")

  (if #t
      (let ((tree (peg:tree (match-pattern root string))))
        (when (< 0 (dzn:debugity))
          (for-each write-symbol (reverse (%peg:table))))
        tree)
      (catch (or (< 0 (dzn:debugity)) 'ignore)
        (lambda _
          (let ((tree (peg:tree (match-pattern root string))))
            (when (< 0 (dzn:debugity))
              (for-each write-symbol (reverse (%peg:table))))
            tree))
        (lambda (key . args)
          (for-each write-symbol (reverse (%peg:table)))))))
